#include <flux/gdb.h>
#include <flux/gdb_serial.h>
#include <flux/tty.h>
#include <flux/x86/trap.h>
#include <flux/machine/base_trap.h>
#include <flux/machine/pc/pic.h>
#include <flux/machine/pc/com_cons.h>

#include <string.h>
#include <setjmp.h>
#include <unistd.h>

#include "thread.h"
#include "region.h"
#include "kmem.h"
#include "lib.h"
#include "config.h"

#include "jdb.h"
#include "kdb.h"

#define ESC 27
#define TRAP_FLAG (1 << 8)
#define RESUME_FLAG (1 << 16)
#define SINGLE_STEP (1 << 14)


jdb::cmd_t jdb::cmd;
unsigned int jdb::pic_mask;
char jdb::show_state = 0, jdb::use_nested = 0;
jdb::breakpoint_t jdb::bps[4] = {{0, }, {0, }, {0, }, {0, }};

char *jdb::mode_names[4] = {"instruction", "write access", 
			    "i/o access", "read/write access"};
char *jdb::debug_msg=NULL;
int jdb::msg_len=0;

unsigned jdb::debug_control;

struct jdb::traps_t jdb::traps[19] = { 
  {"DE", 0}, /* 0 */ 
  {"DB", 0}, 
  {"NI", 0}, 
  {"BP", 0}, 
  {"OF", 0}, 
  {"BR", 0}, /* 5 */ 
  {"UD", 0}, 
  {"NM", 0}, 
  {"DF", 0}, 
  {"OV", 0}, 
  {"TS", 0}, /* 10 */ 
  {"NP", 0}, 
  {"SS", 0}, 
  {"GP", 0}, 
  {"PF", 0}, 
  {"RE", 0}, /* 15 */ 
  {"MF", 0},
  {"AC", 0},
  {"MC", 0}};
 
thread_t *jdb::current_present=0, *jdb::current_ready=0;

int (*jdb::nested_trap_handler)(trap_state *state);

bool jdb::_connected = false;

void
jdb::init()
{
  nested_trap_handler = base_trap_handler;
  base_trap_handler = enter_kdebug;

  _connected = true;
}

bool
jdb::connected()
{
  return _connected;
}

void 
jdb::cursor(int y, int x)
{
  putchar('\006');
  putchar(y);
  putchar(x);
}
void 
jdb::clear_screen()
{
  int i;
  home();
  for(i=0; i<25; i++) {
    putchar('\n');
    clear_to_eol();
  }
  home();
}







void 
jdb::open_debug_console(trap_state *st)
{
  // save pic state and mask all interupts
  debug_control = get_debug_control_register();
  set_debug_control_register(debug_control & ~0xffff00ff);

  unsigned int flags = get_eflags();
  cli();
  pic_mask = inb(MASTER_OCW);
  outb(MASTER_OCW, 0xff);
  set_eflags(flags);

  cursor(23, 0);
  //  printf("L4Kdebug -------------------------------------------------EIP: %08x\n", st->eip);
}
void 
jdb::close_debug_console(void)
{
  com_cons_enable_receive_interrupt();
  pic_enable_irq(config::serial_com_port & 1 ? 4 : 3);
  // restore pic state
  unsigned int flags = get_eflags();
  cli();
  outb(MASTER_OCW, pic_mask);
  set_eflags(flags);

  reset_debug_status_register();
  set_debug_control_register(debug_control);
}


int
jdb::get_32()
{
  int c, val=0;
  for(int digit=0; digit<8; ) {
    c=getchar();
    switch(c) {
    case '0' ... '9':
      val = (val << 4) + (c - '0');
      digit++;
      putchar(c);
      break;
    case 'a' ... 'f':
      val = (val << 4) + (c - 'a' + 10);
      digit++;
      putchar(c);
      break;
    case ' ':
      if (digit)
	return val;
      // fall through
    case ESC:
      return -1;
    }
  }
  return val;
}

int
jdb::get_num(int mask)
{
  int c;
  for(;;) {
    c=getchar();
    if (c == ESC)
      return -1;
    if ((c > '0') && (c <= '9') && (mask & (1 << (c - '1')))) {
      putchar(c);
      return c - '0';
    }
  }
}

l4_threadid_t 
jdb::get_thread_id(void)
{
  int state = 0, num = 0, digit = 0;
  l4_threadid_t id;
  int c;
  for(;;) {
    c=getchar();
    if (state == 0) {
      switch (c) {
      case ' ':
	if (!digit) 
	
  return L4_INVALID_ID; // no id given
	else
	  return L4_NIL_ID; // cancel 
      case ESC : // escape
	  return L4_NIL_ID; // cancel 
      case '.':
	if (digit) {
	  id.id.task = num;
	  num=digit=0;
	  state = 1;
	  putchar(c);
	}
	break;
      case '0' ... '9':
	num = (num << 4) + (c - '0');
	digit++;
	putchar(c);
	break;
      case 'a' ... 'f':
	num = (num << 4) + (c - 'a' + 10);
	digit++;
	putchar(c);
	break;
      }
      if (digit == 3) {
	  id.id.task = num;
	  num=digit=0;
	  state = 1;
	  putchar('.');
      }
    }
    else {
      switch (c) {
      case ' ':
	id.id.lthread = num;
	return id;
      case ESC : // escape
	  return L4_NIL_ID; // cancel 
      case '0' ... '9':
	num = (num << 4) + (c - '0');
	digit++;
	putchar(c);
	break;
      case 'a' ... 'f':
	num = (num << 4) + (c - 'a' + 10);
	digit++;
	putchar(c);
	break;
      }
      if (digit == 2) {
	id.id.lthread = num;
	return id;
      }
    }
  }
}


// thread: 0081 (001.01) <00020401 00080000>	                            prio: 10
// state : 85, ready                            lists: 81                   mcp: ff

// wait for: --                             rcv descr: 00000000   partner: 00000000
// sndq    : 0081 0081                       timeouts: 00000000   waddr0/1: 000/000
// cpu time: 0000000000 timeslice: 01/0a

// pager   : --                            prsent lnk: 0080 0080
// ipreempt: --                            ready link : 0080 0080
// xpreempt: --
//                                         soon wakeup lnk: 
// EAX=00202dfe  ESI=00020401  DS=0008     late wakeup lnk: 
// EBX=00000028  EDI=00080000  ES=0008
// ECX=00000003  EBP=e0020400
// EDX=00000001  ESP=e00207b4

// 700:
// 720:
// 740:
// 760:
// 780:
// 7a0:                                                  0000897b 00000020 00240082
// 7c0:  00000000 00000000 00000000 00000000    00000000 00000000 00000000 00000000
// 7e0:  00000000 00000000 ffffff80 00000000    0000001b 00003200 00000000 00000013
// L4KD: 


static thread_t *get_thread(unsigned long esp)
{
  return reinterpret_cast<thread_t *>(esp & ~(config::thread_block_size - 1));
}

static char *state_names[] = {"ready", "wait", "receiving", "polling",
			      "ipc_in_progress",  "send_in_progress", 
			      "busy", "signalling", "cancel", "dead"};
void 
jdb::show_tcb(l4_threadid_t id, trap_state *st)
{
//   threadid_t dst_id(&dst_id_long);
//   thread_t *dst = (dst_id.lookup() == this)

  int current=1;
  threadid_t t_id(&id);
  thread_t *t = get_thread((unsigned long)st);

  clear_screen();
  home();

  if (t_id.is_valid() && (t != t_id.lookup())) {
    current = 0;
    t = t_id.lookup();
  }

// thread: 0081 (001.01) <00020401 00080000>	                            prio: 10
#if 0
  printf("thread: %3x.%02x <%08x %08x>\tr_prio: %02x prio: %02x\tmcp: %02x\n",
	 t->my_id.id.task, t->my_id.id.lthread, // task.thread
	 t->my_id.lh.high, t->my_id.lh.low, 
	 t->my_real_prio,t->my_prio, t->my_mcp);
#else
  printf("thread: %3x.%02x <%08x %08x>\t\t\tprio: %02x\tmcp: %02x\n",
	 t->my_id.id.task, t->my_id.id.lthread, // task.thread
	 t->my_id.lh.high, t->my_id.lh.low, 
	 t->my_prio, t->my_mcp);
#endif
// state : 85, ready 
  printf("state: %03x ", t->my_state);
  if (t->my_state == 0)
    printf("invalid");
  else {
    unsigned i=0;
    do {
      if (t->my_state & (1<<i))
	printf("%s,", state_names[i]);
      i++;
    } while (i < (sizeof(state_names)/sizeof(char*)));

    putchar('\n');
    // wait for: --                             rcv descr: 00000000   partner: 00000000
    if (t->my_partner)
      {
	if (t->my_partner->my_id.lh.low > 0
	    && t->my_partner->my_id.lh.low <= 16)
	  printf("\nwait for: irq %02x", t->my_partner->my_id.lh.low - 1);
	else
	  printf("\nwait for: %3x.%02x", 
		 t->my_partner->my_id.id.task, 
		 t->my_partner->my_id.id.lthread);
      }
    else
      printf("\nwait for: ---.--");
    if (t->my_send_partner)
      printf("   polling: %3x.%02x", 
	     t->my_send_partner->my_id.id.task, 
	     t->my_send_partner->my_id.id.lthread);
    else
      printf("   polling: ---.--");
    printf("\trcv descr: %08x   partner: %08x\n", 0, 0);
    
    // sndq    : 0081 0081                       timeouts: 00000000   waddr0/1: 000/000
    //printf("sndq\t: %04x %04x\t\t\ttimeouts : %08x   waddr0/1: %03x/%03x\n",
    //	   0, 0, 0, 0, 0);
    if (t->my_signal_partner)
      printf("signalto: %3x.%02x\n", 
	     t->my_signal_partner->my_id.id.task, 
	     t->my_signal_partner->my_id.id.lthread);
    else
      printf("signalto: ---.--\n");
    
    // cpu time: 0000000000 timeslice: 01/0a
    printf("cpu time: %08x timeslice: %02x/%02x\n", 
    	   0, t->my_ticks_left, t->my_timeslice);
    // pager   : --                            prsent lnk: 0080 0080
    if (t->my_pager)
      printf("pager\t: %3x.%02x",
	     t->my_pager->my_id.id.task, t->my_pager->my_id.id.lthread);
    else
      printf("pager\t: ---.--");
    if (t->present_next)
      printf("\t\t\tprsent lnk: %3x.%02x",
	     t->present_next->my_id.id.task, t->present_next->my_id.id.lthread);
    else
      printf("\t\t\tprsent lnk: ---.--");
    if (t->present_prev)
      printf(" %3x.%02x\n", 
	     t->present_prev->my_id.id.task, t->present_prev->my_id.id.lthread);
    else
          printf(" ---.--\n");
    // ipreempt: --                            ready link : 0080 0080
    printf("ipreempt: ---.--\t\t\tready lnk : ");
    if (t->my_state & Thread_running) {
      if (t->ready_next)
	printf("%3x.%02x ",
	     t->ready_next->my_id.id.task, t->ready_next->my_id.id.lthread);
      else
	printf("???.?? ");
      if (t->ready_prev)
	printf("%3x.%02x\n",
	       t->ready_prev->my_id.id.task, t->ready_prev->my_id.id.lthread);
      else
	printf("???.?? ");
    }
    else
      printf("---.-- ---.--\n");
    // xpreempt: --
    printf("xpreempt: ---.--\n");
    
    
    if (current) {
      int from_user = (st->cs & 3);
      
      
      // registers
      printf("EAX=%08x  ESI=%08x  DS=%04x\n",
	     st->eax, st->esi, st->ds);
      printf("EBX=%08x  EDI=%08x  ES=%04x\n", st->ebx, st->edi, st->es);
      printf("ECX=%08x  EBP=%08x\n", st->ecx, st->ebp);
      printf("EDX=%08x  ESP=%08x\n", st->edx, from_user ? st->esp : (unsigned)&st->esp);
      
      printf("trapno %d, error %08x, from %s mode\n",
	     st->trapno, st->err, from_user ? "user" : "kernel");
      printf("CS=%04x  EIP=%08x  EFlags=%08x", st->cs, st->eip, st->eflags);
      if (from_user)
	printf("  SS=%04x  ESP=%08x", st->cs, st->esp);
      putchar('\n');
      if (st->trapno == T_PAGE_FAULT)
	printf("page fault linear address %08x\n", st->cr2);
    }
  }
}

void
jdb::show_ready(trap_state *st)
{
  thread_t *t = current_ready ? current_ready : get_thread((unsigned long)st);
  int c=getchar();
  if (c=='n')
    if (t->ready_next) {
      current_ready = t->ready_next;
      show_tcb(current_ready->my_id, st);
    }
  if (c=='p')
    if (t->ready_prev) {
      current_ready = t->ready_prev;
      show_tcb(current_ready->my_id, st);
    }
}

void
jdb::show_present_or_ptab(trap_state *st)
{
  thread_t *t = current_present ? current_present : get_thread((unsigned long)st);
  int c=getchar();
  if (c=='n')
    if (t->present_next) {
      current_present = t->present_next;
      show_tcb(current_present->my_id, st);
    }
  if (c=='p')
    if (t->present_prev) {
      current_present = t->present_prev;
      show_tcb(current_present->my_id, st);
    }
}

void
jdb::single_step(trap_state *st)
{
  int c;
  printf("\rSinglestep ");
  for(;;) {
    c=getchar();
    switch(c) {
    case '+':
      st->eflags |= TRAP_FLAG;
      printf("enabled");
      break;
    case '-':
      st->eflags &= ~TRAP_FLAG;
      printf("disabled");
      break;
    case 's':
      printf("show state enabled");
      show_state = 1;
      break;
    case 'd':
      printf("show state disabled");
      show_state = 0;
      break;
    case ' ': 
    case ESC: 
      printf("<esc>"); 
      break; 
    default: 
      continue; 
    }
    return; 
  } 
}


#define load_debug_register(val, num) \
    asm("movl %0, %%db" #num ";" : /* no output */ : "r" (val));  

void 
jdb::set_debug_address_register(unsigned long address, int reg_num)
{
  switch (reg_num) {
  case 0:
    load_debug_register(address, 0);
    break;
  case 1:
    load_debug_register(address, 1);
    break;
  case 2:
    load_debug_register(address, 2);
    break;
  case 3:
    load_debug_register(address, 3);
    break;
  default:
    printf("set_debug_register: invalid number\n");
  }
}

void 
jdb::delete_breakpoint()
{
  int i, num=0, used=0;
  for(i=0; (i<4) ; i++)
    if (bps[i].address != 0) {
      num++;
      used=i;
    }
  if (!num) {
    return;
  }
  if (num>1) {
    printf(" number:");
    used=get_num(0xf)-1;
  }
  printf("\ndeleting breakpoint %d\n\n", used+1);
  debug_control &= ~(((3    + (3 << 2)) << (16 + 4*used)) + (3 << (2*used)));
  bps[used].address = 0;
}

void 
jdb::list_breakpoints()
{
  int i;
  putchar('\n');
  for(i=0; (i<4); i++) {
    printf("breakpoint %d: ", i+1);
    if (bps[i].address) {
      printf("break on ");
      switch (bps[i].mode) {
      case 0: 
	printf("instruction at %lx\n", bps[i].address);
	break;
      case 1:
	printf("write at %lx len %d\n", 
	       bps[i].address, bps[i].len);
	break;
      case 2:
	printf("I/O at %lx len %d\n", 
	       bps[i].address, bps[i].len);
	break;
      case 3:
	printf("access at %lx len %d\n", 
	       bps[i].address, bps[i].len);
	break;
      }
    }
    else
      printf("disabled\n");
  }
  printf("\n\n\n");
}

void 
jdb::set_breakpoint(int mode)
{
  int i, len=0;

  for(i=0; (i<4) && (bps[i].address != 0); i++);
  if (i<4) {
    printf(" at: ");
    int address = get_32();
    if (address == -1)
      return;
    if (mode) {
      printf(" len: ");
      len = get_num(0xb)-1; /* 1011 => 4,2,1 */
      if (len == -1)
	return;
    }

    bps[i].address = address;
    bps[i].mode = mode;
    bps[i].len = len+1;
    
    debug_control &= ~(((3    + (3 << 2)) << (16 + 4*i)) + (3 << (2*i)));
    debug_control |=  (((mode + (len<<2)) << (16 + 4*i)) + (2 << (2*i)));

    set_debug_address_register(address, i);

  }
  else {
    printf("No breakpoint available");
  }

  printf("\n\n");
}

void
jdb::break_point(trap_state *st)
{
  int c, mode = 0;
  for(;;) {
    c=getchar();
    switch(c) {
    case 'a':
      mode = 3;
      // access breakpoint, fall through
    case 'w':
      mode |= 1;
      // write breakpoint, fall through
    case 'i':
      // mode == 0, instruction break point
      putchar(c);
      set_breakpoint(mode);
      break;
    case 'p':			// I/O access
      mode = 2;
      putchar(c);
      set_cr4(get_cr4() | CR4_DE);
      set_breakpoint(mode);
      break;
    case '\r':
    case 'l':
      putchar(c);
      list_breakpoints();
      break;
    case '-':
      putchar(c);
      delete_breakpoint();
      break;
    case 'r':
      // read/write/access break point
    case 't':
      // restrict break point to thread
    case 'b':
      // break point base
    case 'x':
      // restrict break point to address range
      putchar(c);
      printf("not implemented yet\n");
      continue;
    default:
      printf("\n");
      break;
    }
    return;
  }
}

void
jdb::show_help(void)
{
  clear_screen();
  home();
  printf("\t?\t\t\tshow help\n"
	 "\t' '\t\t\tshow debug message\n"
	 "\tt[taskno.threadno]\tshow current/given thread\n"
	 "\tr{n|p}\t\t\tshow next/previous thread in ready list\n"
	 "\tp{n|p}\t\t\tshow next/previous thread in present list\n"
	 "\tT{+|-}\t\t\tactivate/deactivate single step\n"
	 "\tb{i|a|w}\t\tset instruction/access/write breakpoint\n"
	 "\tb{l|-}\t\t\tlist/delete breakpoint\n"
	 "\tV{a,h}\t\t\tswitch debugger to vga/hercules screen\n"
	 "\tVs\t\t\tactivate source level debugging\n"
	 "\tg\t\t\tleave kernel debugger\n");
}
extern "C" void switch_vid(int); // in oskit_s_c++.cc

int
jdb::switch_debug_state(trap_state *st)
{
  int c;
  printf("switch_debug_state");
  for(;;) {
    c=getchar();
    switch(c) {
    case 'a':
      switch_vid(0);
      putchar(c);
      return 1;
    case 'h':
      switch_vid(1);
      putchar(c);
      return 1;
    case 's':
      // pass control to source level debugger
      putchar(c);
      if (nested_trap_handler) {
	int r;
	use_nested = 1;
	gdb_trap_recover = 0;
	r = nested_trap_handler(st);
	if (use_nested)
	  return r;      
	return 1;
      }
      else
	printf("no source level debugger available");
    }
  }
}

void 
jdb::page_fault_protocol()
{
  extern bool log_page_fault;

  log_page_fault = (getchar() == '+');
}

void 
jdb::show_kernel_state()
{
  char command = getchar();

  printf("%c\n", command);

  switch (command)
    {
    case 'p':
      {
	// show important I/O ports
	outb_p(MASTER_ICW, OCW_TEMPLATE | READ_NEXT_RD | READ_IS_ONRD);
	unsigned in_service = inb(MASTER_ICW);
	outb_p(MASTER_ICW, OCW_TEMPLATE | READ_NEXT_RD | READ_IR_ONRD);
	unsigned requested = inb(MASTER_ICW);
	printf("master PIC: in service = 0x%x, request = 0x%x, mask = 0x%x\n", 
	       in_service, requested, pic_mask /*(unsigned) inb(MASTER_OCW)*/);
	outb_p(SLAVES_ICW, OCW_TEMPLATE | READ_NEXT_RD | READ_IS_ONRD);
	in_service = inb(SLAVES_ICW);
	outb_p(SLAVES_ICW, OCW_TEMPLATE | READ_NEXT_RD | READ_IR_ONRD);
	requested = inb(SLAVES_ICW);
	printf(" slave PIC: in service = 0x%x, request = 0x%x, mask = 0x%x\n", 
	       in_service, requested, (unsigned) inb(SLAVES_OCW));
      }
      break;

    case 'i':
      {
	space_t *s = (space_t*) kmem::phys_to_virt(get_pdbr());

	printf("clock: %x.%x\n"
	       "pdir:  %x (taskno = %x, chief = %x)\n"
	       , 
	       (unsigned) (kmem::info()->clock >> 32), 
	       (unsigned) (kmem::info()->clock),
	       (unsigned) s, 
	       unsigned(s->space()),
	       unsigned(s->chief()));
      }
      break;

    case 'm':
      kmem::debug_dump();
      break;

    case 'r':
      region::debug_dump();
      break;

    }

  printf("\n");
}

void dummy(void) {
  static int dummy_read, tmp, dummy_write;
  tmp=dummy_read;
  asm("nop;nop;nop;nop;");
  dummy_write=tmp;
}

int 
jdb::execute_command(trap_state *st)
{
  int c;
  for (;;) {
    c=getchar();
    printf("\r\005%c", c);
    switch(c) {
    case 'x':
      show_help();
      break;
    case ' ':
      if (debug_msg) {
	int i=msg_len; 
	char *msg = debug_msg;
	
	putchar('\r');
	for(; i>0; i--)
	  putchar(*msg++);
      }
      break;
    case 't': 
      cmd.command = thread_c;
      cmd.id = get_thread_id();
      if (cmd.id.lh.low != 0) {
	show_tcb(cmd.id, st);
	current_present = current_ready = threadid_t(&cmd.id).lookup();
      }
      break;
    case 'r':
      show_ready(st);
      break;
    case 'p':
      show_present_or_ptab(st);
      break;
    case 'k':
      show_kernel_state();
      break;
    case 'T':
      single_step(st);
      break;
    case 'P':
      page_fault_protocol();
      break;
    case 'b':
      break_point(st);
      break;
    case 'V':
      return switch_debug_state(st);
    case 'g':
      putchar('\n');
      return 0;
    case 'q':
      dummy();
      break;
    case '^':
      exit(1);
    default:
      putchar('\b');
      putchar(' ');
      putchar('\b');
      continue; // ignore character and get next input
    }
    return 1;
  }
}

int
jdb::enter_kdebug(trap_state *st)
{
  static jmp_buf recover_buf;

  static char error_buffer[81] = "";
  
  if (use_nested && nested_trap_handler) {
    int r = nested_trap_handler(st); 
    if (use_nested)
      return r;
  }

  if (gdb_trap_recover)
    longjmp(recover_buf, 1);

  // hack: read serial input to clear irq condition
  if (nested_trap_handler && kdb::connected())
    while (com_cons_try_getchar() != -1) ;

  open_debug_console(st);

  if (st->trapno == 1) {
    int state = get_debug_status_register(), i;
    reset_debug_status_register();
    if (state & SINGLE_STEP) {
      strcpy(error_buffer, "Singlestep");
      if (show_state)
	show_tcb(L4_INVALID_ID, st);
    }
    else {
      for (i=0; i<4; i++) {
	if (state & (1<<i)) {
	  sprintf(error_buffer, "break on %s at %lx", 
		  mode_names[bps[i].mode], bps[i].address);
	  if (!bps[i].mode)
	    st->eflags |=RESUME_FLAG;
	  break;
	}
      }
      if (i==4) {
	sprintf(error_buffer, "unknown debug exception");
      }
    }
  }
  else if (st->trapno == 3) {
    int len = *reinterpret_cast<char*>(st->eip + 1);
    char *msg = reinterpret_cast<char*>(st->eip + 2);
    
    if (*reinterpret_cast<unsigned char*>(st->eip) == 0xeb) {
      if (len > 0) {
	int i;
	len = len < 20 ? len : 20;
	for(i=0; i<len; i++)
	  error_buffer[i] = *(msg+i);
	error_buffer[i]='\0';
      }
    }
    else
      sprintf(error_buffer, "%s", traps[st->trapno].name);
  }
  else if (traps[st->trapno].error_code) {
    sprintf(error_buffer, "%s (%x)", traps[st->trapno].name, st->err);
  }
  else {
    sprintf(error_buffer, "%s", traps[st->trapno].name);
  }

  cursor(24, 0);
  putchar('\n');		// scroll one line

  while (setjmp(recover_buf))
    printf("Invalid access\n\n");

  do {
    cursor(23, 0);
    printf("L4KD -------------------------------------------------EIP: %08x\n", st->eip);
    cursor(23, 6);
    printf("%s", error_buffer);
    cursor(24, 0);
    putchar(5);

    gdb_trap_recover = 1;

  } while (execute_command(st));
  close_debug_console();

  gdb_trap_recover = 0;

  return 0;
}

