#include "thread.h"
#include "globals.h"

#include "lock.h"

// Define this if you want simple locks.  This is unsafe because a
// thread which is killed can take a lock with it.
//#define NO_TRANS_LOCK

unsigned 
thread_t::trans_help()
{
#ifndef NO_TRANS_LOCK
  if (! in_transaction())       // not already in transaction?
    {
      return trans_invoke(this, &trans_help);
    }

  // We are executed in transaction mode.

  // Do nothing.  When this code is executed, this thread will be out
  // of the critical section protected by the lock.  We just rely on
  // the transaction mechanism to switch back to the requester (if he
  // has a higher prio).  (If it wasn't for that switching-back, we
  // wouldn't have to run this transaction at all -- we could just
  // donate our time slice.)

  return signal_t::Sig_success;
#else
  current()->switch_to(this);
  return 0;
#endif
}

inline void 
helping_lock_t::lock_current()
{
  check (current()->switching_in.test_and_set() == false);
}

inline void 
helping_lock_t::unlock_current()
{
  current()->switching_in.clear();
  check (current()->switchin_context() == true);
}

bool 
helping_lock_t::test_and_set()
{
  if (! threading_system_active) // still initializing?
    return false;
#ifndef NO_TRANS_LOCK
  bool current_already_locked = current()->switching_in.test();

  if (! current_already_locked)
    lock_current();
#endif
  if (! compare_and_swap(&_owner, 
			 reinterpret_cast<thread_t*>(0), 
			 current()))
    {
      // Lock was already set.
#if 0 /* the following assert() is not appropriate because we also use
	 this lock to protect against nesting, so that _owner == current(). */
      assert (_owner != current()); 
#endif
#ifndef NO_TRANS_LOCK
      if (! current_already_locked)
	unlock_current();
#endif
      return true;
    }
#ifndef NO_TRANS_LOCK
  _in_trans = current_already_locked;
#endif
  return false;			// We've successfully acquired the lock.
}

void 
helping_lock_t::clear()
{
  if (! threading_system_active) // still initializing?
    return;

  assert(_owner == current());
#ifndef NO_TRANS_LOCK
  bool current_already_locked = _in_trans;
#endif
  _owner = 0;
#ifndef NO_TRANS_LOCK
  if (! current_already_locked)
    unlock_current();
#endif
}
